/*
 * Copyright (C) 2009 Väinö Helminen <vah78@yahoo.com>
 *
 * $Id: conf.c,v 1.1 2009-04-18 05:40:14 vah Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#include <X11/Xutil.h>

#include "conf.h"

/*
 * Trim whitespace from both ends of the string.
 */
static char* trim(char *str)
{
  for(; *str && isspace(*str); ++str) ;
  char *endp;
  for(endp = str; *endp; ++endp);
  for(--endp; endp > str && isspace(*endp); --endp) ;
  endp[1] = '\0';
  return str;
}


int parseConf(const char *filename, Data *data)
{
  FILE *file = fopen(filename, "rb");
  if(!file) {
    perror("evmapper");
    return -1;
  }
  int device = -1, code = -1;
  unsigned command = 0;
  data->devices = 0;
  data->fd = NULL;
  data->mapping = NULL;
  data->command = NULL;

  if(data->verbose) {
    printf("Configuration: %s\n", filename);
  }

  while(1) {
    char buf[1024];
    if(!fgets(buf, 1024, file)) break;
    char *line = trim(buf);
    if(line[0] == '/') {
      if(code == 0) {
	fprintf(stderr, "evmapper: a device without a single mapping defined\n");
	return -1;
      }

      device += 1;
      data->devices = device + 1;
      data->fd = realloc(data->fd, (data->devices) * sizeof *data->fd);
      data->mapping = realloc(data->mapping, (data->devices) * sizeof *data->mapping);
      if(!data->fd || !data->mapping) {
	fprintf(stderr, "evmapper: realloc failed\n");
	return -1;
      }
      data->mapping[device] = NULL;

      if((data->fd[device] = open(line, O_RDONLY)) < 0) {
	perror("evmapper");
	return -1;
      }
      code = 0;

      if(data->verbose) {
	printf("Device: %s\n", line);
      }

    } else if(isdigit(line[0])) {
      if(device < 0 || code < 0) {
	fprintf(stderr, "evmapper: invalid configuration\n");
	return -1;
      }

      data->mapping[device] = realloc(data->mapping[device], (code+1) * sizeof **data->mapping);
      if(!data->mapping[device]) {
	fprintf(stderr, "evmapper: realloc failed\n");
	return -1;
      }

      data->mapping[device][code+1].common.type = 0;

      char *cmdstr = strstr(line, "=\"");
      if(cmdstr) { // command

	cmdstr = trim(cmdstr+2); // skip =" and white space
	
	unsigned eventcode;
	int ret = sscanf(line, "%u =\"", &eventcode);
	if(ret != 1 || !eventcode || !cmdstr[0]) {
	  fprintf(stderr, "evmapper: invalid mapping \"%s\"\n", line);
	  return -1;
	}

	data->command = realloc(data->command, (command+1) * sizeof *data->command);
	if(!data->command) {
	  fprintf(stderr, "evmapper: realloc failed\n");
	  return -1;
	}

	data->mapping[device][code].command.eventcode = eventcode;
	data->mapping[device][code].command.type = COMMAND;
	data->mapping[device][code].command.state = 0;
	data->mapping[device][code].command.command = command;

	if(!(data->command[command] = strdup(cmdstr))) {
	  fprintf(stderr, "evmapper: strdup failed\n");
	  return -1;
	}
	if(data->verbose) {
 	  printf("Mapping: %u =\" %s\n", eventcode, data->command[command]);
	}
	command += 1;

      } else { // key

	unsigned eventcode = 0, modifiers, repeat;
	char keystr[33];
	int ret = sscanf(line, "%u = %32s %u %u", &eventcode, keystr, &modifiers, &repeat);
	if(ret == 2) {
	  modifiers = repeat = 0;
	  if(ret == 2 && keystr[1] == 0 && keystr[0] >= 'A' && keystr[0] <= 'Z') {
	    modifiers = 1;
	  }
	} else if(ret == 3) {
	  repeat = 0;
	}
	unsigned keycode = 0;
	if(ret > 1) {
	  KeySym keysym = XStringToKeysym(keystr);
	  if(keysym != NoSymbol) {
	    keycode = XKeysymToKeycode(data->display, keysym);
	  }
	}
	if(ret < 2 || ret > 4 || !eventcode || !keycode) {
	  fprintf(stderr, "evmapper: invalid mapping \"%s\"\n", line);
	  return -1;
	}
	data->mapping[device][code].key.eventcode = eventcode;
	data->mapping[device][code].key.type = KEY;
	data->mapping[device][code].key.state = 0;
	data->mapping[device][code].key.keycode = keycode;
	data->mapping[device][code].key.modifiers = modifiers;
	data->mapping[device][code].key.repeat = repeat;

	if(data->verbose) {
	  KeySym keysym = XKeycodeToKeysym(data->display, keycode, 0);
	  if(modifiers&1) {
	    KeySym lower;
	    XConvertCase(keysym, &lower, &keysym);
	  }
	  const char *str = XKeysymToString(keysym);
	  printf("Mapping: %u = %s(%u) %u %u\n", eventcode, str, keycode, modifiers, repeat);
	}
      }

      code += 1;
    } // else skip line
  }

  fclose(file);

  if(!data->devices) {
    fprintf(stderr, "evmapper: no devices defined\n");
    return -1;
  }

  return 0;
}
