/*
 * evmapper : map event device scan codes to X keyboard events or external commands
 *
 * Copyright (C) 2009 Väinö Helminen <vah78@yahoo.com>
 *
 * $Id: evmapper.c,v 1.15 2009-04-18 07:20:50 vah Exp $
 */

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "conf.h"
#include "event.h"
#include "x.h"

/*
 * Ignore everything(tm) -error handler for Xlib
 */
int dummyErrorHandler(Display *display, XErrorEvent *event)
{
  display=display; event=event;
  return 0;
}


/*
 * Find the mapping matching the event keycode.
 */
static inline Mapping* findMapping(Mapping *mapping, unsigned eventcode)
{
  if(mapping && eventcode) {
    for(int i = 0; mapping[i].common.type != TERM; ++i) {
      if(mapping[i].common.eventcode == eventcode) {
	return &mapping[i];
      }
    }
  }
  return NULL; // not found
}

/*
 * Send fake key press and release events.
 */
static inline void keyPressRelease(Display *display, unsigned keycode, unsigned modifiers, unsigned verbose)
{
  if(verbose) {
    printf("X key: %u %u\n", keycode, modifiers);
  }
  // try to keep screensaver off
  XResetScreenSaver(display);

  sendKeyEvent(display, keycode, modifiers, 1); // key press
  XFlush(display);
  usleep(100); // keep the key down for a short while
  sendKeyEvent(display, keycode, modifiers, 0); // key release
  XFlush(display);
}

/*
 * Execute a shell command.
 */
static inline void execCommand(const char *command, unsigned verbose)
{
  if(verbose) {
    printf("Execute: %s\n", command);
  }
  if(system(command) < 0) {
    fprintf(stderr, "evmapper: cannot execute \"%s\"\n", command);
  }
}


/*
 * Key event handler
 */
static int handleKey(void *datap, int device, unsigned eventcode, unsigned pressed)
{
  static unsigned prevkey = 0;
  static struct timeval prevtime = {0, 0};

  Data *data = datap;
  if(!data || !data->display || device < 0 || device >= data->devices) {
    fprintf(stderr, "evmapper: huh?\n");
    return -1;
  }

  Mapping *mapping = findMapping(data->mapping[device], eventcode);
  if(!mapping) {
    if(data->verbose) {
      printf("Unmapped: %u\n", eventcode);
    }
    return 0; // ignore unmapped key
  }

  if(data->verbose) {
    printf("%s: %u\n", pressed ? "Press":"Release", eventcode);
  }

  // we only react to key presses, and repeat keys only when asked to do so
  if(!pressed || (mapping->common.state && (mapping->common.type != KEY || !mapping->key.repeat))) {
    mapping->common.state = pressed;
    return 0;
  }
  mapping->common.state = pressed;

  struct timeval now;
  gettimeofday(&now, NULL);

  switch(mapping->common.type) {
  case KEY:
    {
      // limit key repeats
      if(prevkey == eventcode) {
	unsigned repeat = mapping->key.repeat;

	// it is possible to press the same key a few times a second so try not to be too restricting
	if(!repeat) repeat = 3;

	struct timeval diff;
	timersub(&now, &prevtime, &diff);
	if(diff.tv_sec < 1 && diff.tv_usec < 999999L/repeat) return 0;
	prevtime = now;
      }
      keyPressRelease(data->display, mapping->key.keycode, mapping->key.modifiers, data->verbose);
    }
    break;

  case COMMAND:
    {
      struct timeval diff;
      timersub(&now, &prevtime, &diff);
      if(prevkey != eventcode || diff.tv_sec > 0) {
	execCommand(data->command[mapping->command.command], data->verbose);
	prevtime = now;
      }
    }
    break;

  default:
    break; // this should never happen
  }

  prevkey = eventcode;
  return 0;
}


int main(int argc, char **argv)
{
  Data data;
  data.verbose = 0;
  const char *conffile = NULL;

  for(int i = 1; i < argc; ++i) {
    if(argv[i][0] == '-') {
      switch(argv[i][1]) {
      case '-':
	conffile = argv[++i];
	break;
      case 'v':
	++data.verbose;
	break;
      }
    } else if(conffile) {
      conffile = NULL;
      break;
    } else {
      conffile = argv[i];
    }
  }

  if(!conffile) {
    fprintf(stderr, "Usage: %s [-v] configuration\n", argv[0]);
    return EXIT_FAILURE;
  }

  data.display = XOpenDisplay(NULL);
  if(!data.display) {
    fprintf(stderr, "evmapper: Could not open display");
    return EXIT_FAILURE;
  }

  if(parseConf(conffile, &data)) {
    return EXIT_FAILURE;
  }

  XSetErrorHandler(dummyErrorHandler);

  if(data.verbose) {
    puts("Running...");
  }

  int ret = readEvents(data.fd, data.devices, handleKey, &data);
  if(ret == -1) {
    perror("evmapper");
    return EXIT_FAILURE;
  } else if(ret) {
    fprintf(stderr, "evmapper: Oops!\n");
    return EXIT_FAILURE;
  }

  XCloseDisplay(data.display);

  return EXIT_SUCCESS;
}
